# -*- coding: utf-8 -*-
import os

working_settings = {}

DEFAULTS = {
    'PROJECT_NAME': os.environ.get('PROJECT_NAME', ''),
    'PROJECT_VERSION': os.environ.get('PROJECT_VERSION', ''),
    'HTTP_PORT': 80,
    'WORKERS': 1,
    'INSPECTOR_REPORT_INTERVAL': 30,

    'DEV': False,
    'CORS': False,
    'DEBUG': False,

    'LOG_LEVEL': {
        'other': 'WARNING',
        'wings_sanic': 'INFO',
        'project': 'INFO'
    },
    'GLOBAL_URL_PREFIX': '',
    'BLUEPRINTS': [],
    'DEFAULT_CONTEXT': {
        'response_shape': 'wings_sanic.views.ResponseShape',
        'serialize_when_none': True
    },
    'EXCEPTION_HANDLER': 'wings_sanic.views.exception_handler',

    'SWAGGER': {
        'info': {
            "version": os.environ.get('PROJECT_VERSION', '1.0.0'),
            "title": os.environ.get('PROJECT_NAME', 'API'),
            "description": '',
            "termsOfService": None,
            "contact": {
                "email": None
            },
            "license": {
                "email": None,
                "url": None
            }
        },
        'schemes': ['http']
    },
    'MQ_SERVERS': {
        # 'default': {
        #     'server': 'wings_sanic.events.rabbitmq_server.MqServer',
        #     'url': 'amqp://guest:guest@127.0.0.1:5672',
        #     'exchange': 'test',
        #     'exchange_type': 'topic',
        #     'reconnect_delay': 5.0
        # }
    },
    'EVENT_HANDLE_TIMEOUT': 10,
    'EVENT_MAX_RETRY': -1,

    # rpc， mq传递上下文信息时传递指定的key, 加入headers时，
    # 1. 自动加X-前缀
    # 2. 转化为大些
    # 3. _转化为-
    'CONTEXT_WHEN_DELIVERY': []
}


def load(**user_settings):
    working_settings.update(user_settings)


def get(attr):
    try:
        # Check if present in user settings
        return working_settings[attr]
    except KeyError:
        # default LOG(JsonFormatter) is useful for elk, but not friendly for develop debug.
        # for convenience for develop, set True, then use sanic's log and its access_log is available
        if attr == 'LOGGING_CONFIG':
            return {
                'version': 1,
                'disable_existing_loggers': False,
                'loggers': {
                    "": {
                        "level": get('LOG_LEVEL').get('other', 'WARNING'),
                        "handlers": ["default"],
                        'propagate': False,
                    },
                    'wings_sanic': {
                        "level": get('LOG_LEVEL').get('wings_sanic', 'INFO'),
                        "handlers": ["wings_sanic"],
                        'propagate': False,
                    },
                    'project': {
                        "level": get('LOG_LEVEL').get('project', 'INFO'),
                        "handlers": ["project"],
                        'propagate': False,
                    }
                },
                'handlers': {
                    "default": {
                        "class": "logging.StreamHandler",
                        "formatter": "json",
                        "level": get('LOG_LEVEL').get('other', 'WARNING'),
                    },
                    "wings_sanic": {
                        "class": "logging.StreamHandler",
                        "formatter": "json",
                        "level": get('LOG_LEVEL').get('wings_sanic', 'INFO'),
                    },
                    "project": {
                        "class": "logging.StreamHandler",
                        "formatter": "json",
                        "level": get('LOG_LEVEL').get('project', 'INFO'),
                    }
                },
                'formatters': {
                    "json": get('DEV') and
                            {
                                "format": "%(asctime)s [%(process)d] [%(levelname)s] %(message)s",
                                "datefmt": "[%Y-%m-%d %H:%M:%S %z]",
                                "class": "logging.Formatter",
                            } or
                            {
                                "class": "wings_sanic.log_formatter.JsonFormatter"
                            }
                }
            }
        try:
            return DEFAULTS[attr]
        except KeyError:
            raise AttributeError("Invalid setting: '%s'" % attr)


def __getattr__(name):
    return get(name)



import copy
import inspect
import os
import re
from itertools import repeat

from sanic.blueprints import Blueprint
from sanic.response import json
from sanic.views import CompositionView

from wings_sanic import settings, utils, serializers
from wings_sanic.views import get_response_shape

blueprint = Blueprint('swagger', url_prefix='/swagger')

_spec = {}


def __summary_description(doc_string):
    doc_string = (doc_string or "").strip()
    s = doc_string.split('\n', 1)
    if len(s) == 1:
        s.append("")
    return s[0].strip(), s[1].strip()


def __update_definitions(group_spec, serializer):
    # 解析serializer
    if isinstance(serializer, serializers.BaseSerializer):
        serializer = serializer.child if isinstance(serializer, serializers.ListSerializer) else serializer
        cls_str = utils.cls_str_of_obj(serializer)
        if cls_str in serializers.definitions:
            group_spec['definitions'][cls_str] = serializers.definitions[cls_str]
        # 递归解析serializer的fields(其中可能存在SerializerField)
        for field in serializer.fields.values():
            __update_definitions(group_spec, field)
    # 解析field
    if isinstance(serializer, serializers.SerializerField):
        __update_definitions(group_spec, serializer.serializer)
    if isinstance(serializer, serializers.ListField):
        __update_definitions(group_spec, serializer.field)


@blueprint.listener('before_server_start')
def build_spec(app, loop):
    for uri, route in app.router.routes_all.items():
        if uri.startswith(f"{settings.GLOBAL_URL_PREFIX}/swagger"):
            continue

        # Build list of methods and their handler functions
        handler_type = type(route.handler)
        if handler_type is CompositionView:
            method_handlers = route.handler.handlers.items()
        else:
            method_handlers = zip(route.methods, repeat(route.handler))

        # format path parameters '<param>' to '{param}'
        uri_parsed = uri
        for parameter in route.parameters:
            uri_parsed = re.sub('<' + parameter.name + '.*?>', '{' + parameter.name + '}', uri_parsed)

        for _method, _handler in method_handlers:
            metadata = utils.get_value(_handler, 'metadata')
            if _method == 'OPTIONS' or not metadata or metadata.swagger_exclude:
                continue

            # ensure group info
            group = metadata.swagger_group or 'default'
            if not isinstance(group, dict):
                group = {'title': str(group)}
            group_title = group.get('title', None) or 'default'
            if group_title not in _spec:
                _spec[group_title] = copy.deepcopy(settings.get('SWAGGER'))
                _spec[group_title].update({
                    'swagger': '2.0',
                    'definitions': {},
                    'paths': {},
                })
                _spec[group_title]['info'].update(group)

            group_spec = _spec[group_title]
            if uri_parsed not in group_spec['paths']:
                group_spec['paths'][uri_parsed] = {}

            # generate path
            parameters = []
            # header
            for name, field in (utils.get_value(metadata.header_serializer, 'fields') or {}).items():
                parameter = field.openapi_spec()
                parameter.update({'in': 'header'})
                parameters.append(parameter)
                __update_definitions(group_spec, metadata.header_serializer)

            # path
            for name, field in (utils.get_value(metadata.path_serializer, 'fields') or {}).items():
                parameter = field.openapi_spec()
                parameter.update({'in': 'path', 'required': True})
                parameters.append(parameter)
                __update_definitions(group_spec, metadata.path_serializer)

            # query
            for name, field in (utils.get_value(metadata.query_serializer, 'fields') or {}).items():
                parameter = field.openapi_spec()
                parameter.update({'in': 'query'})
                parameters.append(parameter)
                __update_definitions(group_spec, metadata.query_serializer)

            # body
            has_file_filed = False
            if metadata.body_serializer:
                body_spec = metadata.body_serializer.openapi_spec()
                import wings_sanic
                for _, f in metadata.body_serializer.fields.items():
                    if isinstance(f, wings_sanic.FileField):
                        has_file_filed = True
                        break

                if not has_file_filed:
                    parameters.append({
                        'in': 'body',
                        'name': 'body',
                        'required': True,
                        'schema': body_spec
                    })
                else:
                    for name, field_spec in body_spec['properties'].items():
                        parameters.append({
                            'in': 'formData',
                            **field_spec
                        })

                __update_definitions(group_spec, metadata.body_serializer)

            # response
            response_spec = None
            if metadata.response_serializer:
                response_spec = metadata.response_serializer.openapi_spec()

                __update_definitions(group_spec, metadata.response_serializer)

            response_shape = get_response_shape(metadata.context)
            response_spec = response_shape.swagger(response_spec)

            summary, description = __summary_description(inspect.cleandoc(_handler.__doc__ or ""))

            endpoint = {
                'operationId': utils.meth_str(_handler),
                'summary': summary,
                'description': description,
                'consumes': ['multipart/form-data'] if has_file_filed else ['application/json'],
                'produces': ['application/json'],
                'tags': metadata.tags,
                'parameters': parameters,
                'responses': {
                    "200": {
                        "description": None,
                        "examples": None,
                        "schema": response_spec
                    }
                },
            }

            group_spec['paths'][uri_parsed][_method.lower()] = endpoint


@blueprint.route('/group/')
def spec_group(request, *args, **kwargs):
    # default第一个，其他是排序后的结果
    groups = [{'name': title, 'url': f'{settings.GLOBAL_URL_PREFIX}/swagger/openapi/?group={title}'}
              for title, _ in _spec.items() if title != 'default']
    groups.sort(key=lambda i: i['name'])
    groups.insert(0, {'name': 'default', 'url': f'{settings.GLOBAL_URL_PREFIX}/swagger/openapi/?group=default'})
    return json(groups)


@blueprint.route('/openapi/')
def spec(request, *args, **kwargs):
    title = utils.get_value(request.raw_args, 'group') or 'default'
    return json(_spec[title])


dir_path = os.path.dirname(os.path.realpath(__file__))
blueprint.static('/', dir_path)
blueprint.static('//', dir_path + '/index.html')



# -*- coding: utf-8 -*-
import asyncio
import json
import logging

from . import datetime_helper, registry, context_var, settings, utils

logger = logging.getLogger(__name__)


class DomainEvent:
    def __init__(self, event_name, **kwargs):
        if 'version' in kwargs:
            kwargs['version'] = kwargs['version'] + 1
        self.occur_on = datetime_helper.get_utc_time()
        self.event_name = event_name
        self.__dict__.update(kwargs)


def __get_mq_server(server_name):
    server = registry.get(server_name, 'mq_servers')
    if not server:
        raise Exception('cannot find named "{0}" mq_server'.format(server_name))
    return server


async def publish(event: DomainEvent, mq_server='default', send_after_done=True):
    """
    发送事件
    """
    await publish_message(event.event_name, event, mq_server, send_after_done)


async def publish_message(routing_key: str, message, mq_server='default', send_after_done=True):
    """
    发送消息
    """
    ctx = context_var.get() or {}
    # 默认必传递的信息
    ctx_delivery = {
        'X-TRACE-ID': ctx.get('trace_id', ''),
        'X-EVENT-NAME': routing_key,
        'X-OCCUR-ON': datetime_helper.now()
    }
    # 用户定制传递的消息头
    for key in settings.get('CONTEXT_WHEN_DELIVERY'):
        ctx_delivery['X-' + key.upper().replace('_', '-')] = utils.get_value(ctx, key)

    body = {
        'payload': message,
        'headers': ctx_delivery
    }
    body = json.dumps(utils.to_primitive(body))
    server = __get_mq_server(mq_server)

    # 立即发送
    # 1. context_var未设置，无法收集message
    # 2. 调用设置send_after_done=False
    if not ctx or not send_after_done:
        await server.publish(routing_key, body)
    else:
        messages = ctx.get('messages', [])
        messages.append({'mq_server': server, 'body': body, 'routing_key': routing_key})
        ctx['messages'] = messages


def handler(event_name, mq_server='default', msg_type=DomainEvent, timeout=None, max_retry=None, subscribe=False):
    """
    添加事件处理器的装饰器
    :param event_name: 事件名
    :param mq_server: mq_server名字
    :param msg_type: 将接受到的内容转化成该msg_type对应的类型，如果为None或者转化不成功则返回原字符串
    :param timeout: 处理的超时时长，默认值10秒，优先级 默认值 < settings设置 < 装饰器设置
    :param max_retry: 最大重试次数，默认值-1(无限重试)，优先级 默认值 < settings设置 < 装饰器设置
    :param subscribe: True是pub-sub模式, 否则producer-consumer模式
    :return: 
    """
    timeout = timeout if timeout is not None else settings.get('EVENT_HANDLE_TIMEOUT')
    timeout = timeout if timeout is not None else 10
    max_retry = max_retry if max_retry is not None else settings.get('EVENT_MAX_RETRY')
    max_retry = max_retry if max_retry is not None else -1

    def decorator(func):
        # 将handler注册到registry，app启动后初始化订阅
        handlers_cur = registry.get(mq_server, 'event_handlers') or set()
        handlers_cur.add((event_name, func, msg_type, timeout, max_retry, subscribe))
        registry.set(mq_server, handlers_cur, 'event_handlers')

        return func

    return decorator


def commit_events():
    try:
        messages = utils.get_value(context_var.get(), 'messages', [])
        for i in messages:
            server = i['mq_server']
            asyncio.run_coroutine_threadsafe(server.publish(i['routing_key'], i['body']), server.loop)
    except Exception as ex:
        logger.error('send event error', exc_info=ex)
